"""Simple box-like environments."""

from dataclasses import dataclass
from os import PathLike

from torchradio.device import Position
from torchradio.env.null import NullEnvironment


@dataclass
class Bounds3D:
    """A 3D region bounded by `[0, 0, 0]` and `[x_max, y_max, z_max]`.

    Args:
        x_max: Bound x coordinates between 0 and `x_max`.
        y_max: Bound y coordinates between 0 and `y_max`.
        z_max: Bound z coordinates between 0 and `z_max`.

    Example:
        ```
        >>> bounds = Bounds3D(10, 10, 5)
        >>> Position(9, 6, 4) in bounds
        True
        >>> Position(9, 6, 6) in bounds
        False
        >>> Position(9, -1, 4) in bounds
        False
        ```

    """

    x_max: float
    y_max: float
    z_max: float

    def __contains__(self, position: Position) -> bool:
        """Determine if position is inside `Bounds3D`.

        Args:
            position: Position to test.

        Returns:
            True if the position is in bounds.

        """
        if position.x < 0 or position.y < 0 or position.z < 0:
            return False

        return (
            position.x <= self.x_max
            and position.y <= self.y_max
            and position.z <= self.z_max
        )


class BoxEnvironment(NullEnvironment):
    """A simple environment defined by a 3D region."""

    def __init__(self, x_max: float, y_max: float, z_max: float) -> None:
        """Create a new bounded `BoxEnvironment`.

        A `Position` is in bounds if `0 <= x <= x_max`, `0 <= y <= y_max` and `0 <= z <= z_max`.

        Args:
            x_max: Maximum x coordinate
            y_max: Maximum y coordinate
            z_max: Maximum z coordinate


        Raises:
            ValueError: if `x_max < 0` or `y_max < 0` or `z_max < 0`.

        """
        if x_max < 0:
            err = f"{x_max=} < 0"
            raise ValueError(err)

        if y_max < 0:
            err = f"{y_max=} < 0"
            raise ValueError(err)

        if z_max < 0:
            err = f"{z_max=} < 0"
            raise ValueError(err)

        self._bounds = Bounds3D(x_max, y_max, z_max)
        super().__init__()

    @property
    def bounds(self) -> Bounds3D:
        """Public API for internal bounds.

        Returns
            Three-dimensional bounds for the environment.

        """
        return self._bounds

    def visualize(
        self,
        *,
        show: bool = True,
        save_path: PathLike | None = None,
    ) -> None:
        """Visualize the environment with the currently placed devices."""
        raise NotImplementedError

    def _in_bounds(self, position: Position) -> bool:
        return position in self._bounds


class PlanarEnvironment(BoxEnvironment):
    """A simple environment defined by a flat 2D region.

    Devices must have spatial distributions that place them at a height of z=0.
    """

    def __init__(self, x_max: float, y_max: float) -> None:
        """Create a new bounded `PlanarEnvironment`.

        A `Position` is in bounds if `0 <= x <= x_max`, `0 <= y <= y_max` and `z == 0`.

        Args:
            x_max: Maximum x coordinate
            y_max: Maximum y coordinate


        Raises:
            ValueError: if `x_max < 0` or `y_max < 0`.

        """
        super().__init__(x_max, y_max, 0)

    def visualize(
        self,
        *,
        show: bool = True,
        save_path: PathLike | None = None,
    ) -> None:
        """Visualize the environment with the currently placed devices."""
        raise NotImplementedError



"""Example trainable algorithms to get started with.

New users are recommended to use `DenseRadio` to ensure that input parameters are compatible.
A `DenseRadio` provides compatible `tx` and `rx` methods. When defining transmitters and receivers,
ensure the corresponding method is passed.

"""

import warnings

import torch
from torch import nn

from torchradio import Reception, Transmission
from torchradio.algorithm.utils import get_random_bits
from torchradio.types import COMPLEX_DTYPE


class DenseTransmissionAlgorithm(nn.Module):
    """An opinionated `TransmissionAlgorithm` that uses trainable dense layers.

    Do not expect this algorithm to be especially performant. It is an example to help
    users get started with training transmitters using `Torchradio`.
    """

    def __init__(self, n_input_bits: int, tx_length_per_bit: float) -> None:
        """Create a new `DenseTranmissionAlgorithm`.

        Args:
            n_input_bits: An input parameter used to shape the number of parameters in the
                dense layers.
            tx_length_per_bit: An input parameter used to shape the number of parameters in the
                dense layers.

        """
        super().__init__()

        self._input_length = n_input_bits
        self._output_length = int(tx_length_per_bit * n_input_bits)

        # don't worry new users about Lazy API warnings
        warnings.filterwarnings("ignore", module="torch")

        self._net = nn.Sequential(
            nn.LazyLinear(4 * self._output_length),
            nn.ReLU(),
            nn.LazyLinear(2 * self._output_length),
        )

    def __call__(self, n_timesteps: int, batch_size: int = 1) -> Transmission:
        """Apply a series of dense layers to randomly generated bits.

        Args:
            n_timesteps: How many timesteps to transmit for.
            batch_size: How many transmissions to create in parallel.

        Returns:
            A `Transmission` with a complex-valued signal tensor with shape `[n_timesteps, batch_size]`
                and a `metadata` dictionary that records the transmitted bits using the key "bits".

        """
        transmission = torch.zeros([batch_size, n_timesteps], dtype=COMPLEX_DTYPE)
        bits = get_random_bits(
            self._input_length * (n_timesteps // self._output_length),
            batch_size,
        )
        signal = self._net(bits.reshape(-1, self._input_length).float()).reshape(
            batch_size,
            -1,
            2,
        )
        transmission[:, : len(signal[0])] += signal[..., 0] + 1j * signal[..., 1]

        return Transmission(
            signal=transmission,
            metadata={"bits": bits},
        )


class DenseReceptionAlgorithm(nn.Module):
    """An opinionated `ReceptionAlgorithm` that uses trainable dense layers.

    Do not expect this algorithm to be especially performant. It is an example to help
    users get started with training receivers using `Torchradio`.
    """

    def __init__(self, n_bits: int, window_length: int) -> None:
        """Create a new `DenseReceptionAlgorithm`.

        Args:
            n_bits: An input parameter used to shape the number of parameters in the
                dense layers. Should be compatible with the `n_input_bits` specified in
                `DenseTransmissionAlgorithm`.
            window_length: An input parameter used cutoff unused bits from
                `DenseTransmissionAlgorithm`. Usually this is set to
                `n_input_bits * tx_length_per_bit` from the instantiation of
                `DenseTransmissionAlgorithm`.


        """
        super().__init__()

        self._n_bits = n_bits
        self._window_length = window_length

        # don't worry new users about Lazy API warnings
        warnings.filterwarnings("ignore", module="torch")

        self._net = nn.Sequential(
            nn.LazyLinear(4 * self._n_bits),
            nn.ReLU(),
            nn.LazyLinear(2 * self._n_bits),
            nn.ReLU(),
            nn.LazyLinear(self._n_bits),
            nn.Sigmoid(),
        )

    def __call__(self, signal: torch.Tensor) -> Reception:
        """Apply a series of dense layers an input signal.

        Args:
            signal: A 2D complex valued input tensor.


        Returns:
            A `Reception` dictionary that records the decoded bits using the key "bits",
                and the probability of each bit under "bit_probabilities".

        """
        batch_size = signal.shape[0]

        # trim unused timesteps
        to_trim = signal.shape[1] % self._window_length
        if to_trim != 0:
            signal = signal[:, :-to_trim]

        reshaped = (
            torch.stack([signal.real, signal.imag])
            .transpose(0, 1)
            .transpose(2, 1)
            .reshape(-1, 2 * self._window_length)
        )
        bit_probabilities = self._net(reshaped).reshape(batch_size, -1)

        decoded = torch.round(bit_probabilities).int()

        return {
            "bit_probabilities": bit_probabilities,
            "bits": decoded,
        }


class DenseRadio(nn.Module):
    """A convenient wrapper for `DenseTransmissionAlgorithm` and `DenseReceptionAlgorithm`.

    `DenseRadio` guarantees that the underlying `TransmissionAlgorithm` and `ReceptionAlgorithm` are
    compatible. A `DenseRadio` provides compatible `tx` and `rx` methods. When defining transmitters
    and receivers, ensure the corresponding method is passed.

    Example:
        ```
        >>> radio = DenseRadio(8, 2)
        >>> env = NullEnvironment()
        >>> env.place({"tx": radio.tx}, {"rx": radio.rx})
        ```

    """

    def __init__(self, n_input_bits: int, tx_length_per_bit: float) -> None:
        """Create a new `DenseRadio`. See DenseTransmissionAlgorithm.__init__."""
        super().__init__()

        self._tx = DenseTransmissionAlgorithm(n_input_bits, tx_length_per_bit)
        self._rx = DenseReceptionAlgorithm(
            n_input_bits,
            int(n_input_bits * tx_length_per_bit),
        )

    def tx(self, n_timesteps: int, batch_size: int = 1) -> Transmission:
        """See `DenseTransmissionAlgorithm.__call__`."""
        return self._tx(n_timesteps, batch_size)

    def rx(self, signal: torch.Tensor) -> Reception:
        """See `DenseReceptionAlgorithm.__call__`."""
        return self._rx(signal)



"""Subpackage for defining common simulation environments."""

from torchradio.env.base import BaseEnvironment
from torchradio.env.box import (
    BoxEnvironment,
    PlanarEnvironment,
)
from torchradio.env.null import NullEnvironment

__all__ = [
    "BaseEnvironment",
    "NullEnvironment",
    "BoxEnvironment",
    "PlanarEnvironment",
]



"""A device pairs a `TransmissionAlgorithm` or a `ReceptionAlgorithm` with other physical characteristics."""

from collections.abc import Iterator

import torch

from torchradio.position import Position, SpatialDistribution, get_null_distribution
from torchradio.types import (
    Reception,
    ReceptionAlgorithm,
    Transmission,
    TransmissionAlgorithm,
)


class Device:
    """The base `Device` class consists of a `SpatialDistribution` that is sampled whenever the device is placed in an environment."""

    def __init__(
        self,
        spatial_distribution: SpatialDistribution | None = None,
    ) -> None:
        """Create a new device with a `spatial_distribution`.

        Args:
            spatial_distribution: A `Callable` that randomly samples `Position`s.
                Defaults to None. If None, a null `SpatialDistribution` is used,
                which always returns `Position(x=0, y=0, z=0)`.

        """
        if spatial_distribution is None:
            spatial_distribution = get_null_distribution()

        self._spatial_distribution = spatial_distribution

    @property
    def spatial_distribution(self) -> SpatialDistribution:
        """Get the device's spatial distribution.

        Returns
            The device's spatial distribution.

        """
        return self._spatial_distribution

    def place(self) -> Position:
        """Sample a `Position` from the device's spatial distribution.

        Returns
            A `Position` randomly sampled according to the device's spatial distribution.

        """
        return self._spatial_distribution()

    @staticmethod
    def _get_parameters(
        algorithm: TransmissionAlgorithm | ReceptionAlgorithm,
    ) -> Iterator[torch.nn.Parameter]:
        """Get the trainable parameters for an algortihm.

        Returns
            The algorithms's trainable parameters. An empty iterator is returned if the
                algortihm does not have trainable parameters.

        """
        if isinstance(algorithm, torch.nn.Module):
            return algorithm.parameters()
        return iter(())


class Transmitter(Device):
    """A `Device` that transmits signals into the environment."""

    def __init__(
        self,
        algorithm: TransmissionAlgorithm,
        spatial_distribution: SpatialDistribution | None = None,
        max_gain: float | None = None,
    ) -> None:
        """Create a new `Transmitter` from an algorithm and a spatial distribution.

        Args:
            algorithm: A `Callable` that produces `Transmission`s.
            spatial_distribution: A `Callable` that randomly samples `Position`s.
                Defaults to None. If None, a null `SpatialDistribution` is used,
                which always returns `Position(x=0, y=0, z=0)`.
            max_gain: A saturation limit on transmissions. The real and imaginary
                components of transmitted signals are capped to `max_gain`.

        """
        super().__init__(spatial_distribution)
        self._algorithm = algorithm
        self._max_gain = max_gain

    def __call__(self, n_timesteps: int, batch_size: int = 1) -> Transmission:
        """Invoke the underlying `TransmissionAlgorithm` to produce a `Tranmission`.

        Args:
            n_timesteps: Number of timesteps to transmit for.
            batch_size: How many simulations to conduct in parallel.

        Returns:
            A `Transmission` with a complex-valued signal with shape `[batch_size, n_timesteps]`.

        """
        if n_timesteps < 1:
            err = f"{n_timesteps=} < 1"
            raise ValueError(err)

        tx = self._algorithm(n_timesteps, batch_size)

        if "raw" in tx.metadata:
            err = f'{self._algorithm} uses the reserved keyword "raw" in the output metadata dictionary.'
            raise RuntimeError(err)

        tx.metadata["_raw"] = tx.signal  # track desired output signal under "raw"

        # cap any real or imaginary values to self._max_gain
        if (
            self._max_gain is not None
            and torch.max(torch.abs(tx.signal)) > self._max_gain
        ):
            real = tx.signal.real
            imag = tx.signal.imag
            real[real > self._max_gain] = self._max_gain
            imag[imag > self._max_gain] = self._max_gain
            tx.signal = torch.complex(real, imag)

        return tx

    def parameters(self) -> Iterator[torch.nn.Parameter]:
        """Get the transmitter's trainable parameters.

        Returns
            The transmitter's trainable parameters. An empty iterator is returned if the
                transmitter does not have trainable parameters.

        """
        return self._get_parameters(self._algorithm)


class Receiver(Device):
    """A `Device` that receives signals from the environment."""

    def __init__(
        self,
        algorithm: ReceptionAlgorithm,
        spatial_distribution: SpatialDistribution | None = None,
    ) -> None:
        """Create a new `Receiver` from an algorithm and a spatial distribution.

        Args:
            algorithm: A `Callable` that produces `Receptions`s.
            spatial_distribution: A `Callable` that randomly samples `Position`s.
                Defaults to None. If None, a null `SpatialDistribution` is used,
                which always returns `Position(x=0, y=0, z=0)`.

        """
        super().__init__(spatial_distribution)
        self._algorithm = algorithm

    def __call__(self, signal: torch.Tensor) -> Reception:
        """Invoke the underlying `ReceptionAlgorithm` to produce a `Reception` dictionary.

        Args:
            signal: Input signal from the environment.

        Returns:
            A dictionary the summarizes various aspects of the received signal. Common
                keys for reconstructive receivers include 'bit_probabilities' and 'bits'.

        """
        return self._algorithm(signal)

    def parameters(self) -> Iterator[torch.nn.Parameter]:
        """Get the receiver's trainable parameters.

        Returns
            The receiver's trainable parameters. An empty iterator is returned if the
                receiever does not have trainable parameters.

        """
        return self._get_parameters(self._algorithm)



import pytest
import torch

from torchradio import Receiver, Transmitter
from torchradio.algorithm.null import (
    get_constant_transmission_algorithm,
    get_null_reception_algorithm,
)
from torchradio.env.base import BaseEnvironment
from torchradio.env.box import (
    BoxEnvironment,
    PlanarEnvironment,
)
from torchradio.env.null import (
    ControlledSNREnvironment,
    NullEnvironment,
    RandomAWGNEnvironment,
)


class SimpleNonDifferentiableEnvironment(NullEnvironment):
    """An example non-differentiable environment to test that `.is_differentiable()` returns False."""

    def _propagate(self, tx: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        # The same `_propagate` method as NullEnvironment, but the received tensors
        # are detached to break differentiability.
        summed = torch.sum(torch.concat(list(tx.values())), axis=0)  # type: ignore
        return {receiver_name: summed.detach() for receiver_name in self._receivers}


class NewParamNonDifferentiableEnvironment(NullEnvironment):
    """An example non-differentiable environment to test that `.is_differentiable()` returns False."""

    def _propagate(self, tx: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        # The same `_propagate` method as NullEnvironment, but the received tensors
        # are detached and added to a new parameter to break differentiability.
        summed = torch.sum(torch.concat(list(tx.values())), axis=0)  # type: ignore
        new_param = torch.nn.Parameter(torch.tensor(0 + 0j))
        return {
            receiver_name: new_param + summed.detach()
            for receiver_name in self._receivers
        }


@pytest.fixture
def null_environment() -> NullEnvironment:  # noqa: D103
    return NullEnvironment()


@pytest.mark.parametrize(
    "env",
    [
        NullEnvironment(),
        RandomAWGNEnvironment(0.1, 0.1),
        ControlledSNREnvironment(10),
        BoxEnvironment(10, 10, 10),
        PlanarEnvironment(10, 10),
    ],
)
def test_differentiability(env: BaseEnvironment) -> None:
    """Check that all Torchradio environments are differentiable."""
    assert env.is_differentiable()


@pytest.mark.parametrize(
    "env_cls",
    [
        SimpleNonDifferentiableEnvironment,
        NewParamNonDifferentiableEnvironment,
    ],
)
def test_non_differentiability(env_cls: type[BaseEnvironment]) -> None:
    """Check test non-differentiable environments are actually non-differentiable.

    Assert that a non-differentiability warning is raised upon initialization.

    """
    with pytest.warns(UserWarning):
        env = env_cls()
    assert not env.is_differentiable()


@pytest.mark.parametrize(
    ("transmitters", "expected_value"),
    [
        ({"tx": Transmitter(get_constant_transmission_algorithm(1 + 1j))}, 1 + 1j),
        (
            {
                "tx1": Transmitter(get_constant_transmission_algorithm(1 + 1j)),
                "tx2": Transmitter(get_constant_transmission_algorithm(2 + 2j)),
            },
            3 + 3j,
        ),
    ],
)
def test_null_correct(
    null_environment: NullEnvironment,
    transmitters: dict[str, Transmitter],
    expected_value: complex,
) -> None:
    """Check that NullEnvironment has been implemented correctly using constant transmissions."""
    receiver = Receiver(get_null_reception_algorithm())

    # place and simulate
    null_environment.place(transmitters, {"rx1": receiver, "rx2": receiver})
    device_logs = null_environment.simulate(10, 2)

    assert torch.all(device_logs.rx["rx1"]["raw"] == expected_value).item()
    assert torch.all(device_logs.rx["rx2"]["raw"] == expected_value).item()



"""Simple (null) transmitters for testing and debugging."""

import torch

from torchradio import (
    Reception,
    ReceptionAlgorithm,
    Transmission,
    TransmissionAlgorithm,
)
from torchradio.types import COMPLEX_DTYPE


def get_constant_transmission_algorithm(
    val: complex | torch.nn.Parameter,
) -> TransmissionAlgorithm:
    """Get a basic transmission algorithm that output a constant complex value.

    Args:
        val: constant complex value to transmit

    Returns:
        A `TransmissionAlgorithm` that outputs a tensor where every element is `val`.

    Example:
        ```
        >>> tx = get_constant_transmitter(1 + 1j)
        >>> tx(5, 2)
        Transmission(signal=tensor([[1.+1.j, 1.+1.j, 1.+1.j, 1.+1.j, 1.+1.j],
                [1.+1.j, 1.+1.j, 1.+1.j, 1.+1.j, 1.+1.j]]), metadata={})
        ```

    """

    def transmission_algorithm(n_timesteps: int, batch_size: int = 1) -> Transmission:
        vector = val * torch.ones(batch_size, n_timesteps, dtype=COMPLEX_DTYPE)
        return Transmission(signal=vector, metadata={})

    return transmission_algorithm


def get_null_transmission_algortihm() -> TransmissionAlgorithm:
    """Get a null transmission algorithm that output nothing.

    Returns:
        A `TransmissionAlgorithm` that outputs a complex-valued tensors of zeros.

    Example:
        ```
        >>> tx = get_null_transmission_algortihm()
        >>> tx(5, 2)
        Transmission(signal=tensor([[0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
                [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j]]), metadata={})
        ```

    """
    return get_constant_transmission_algorithm(0 + 0j)


def get_null_reception_algorithm() -> ReceptionAlgorithm:
    """Get a null reception algorithm that does not do any processing,.

    Returns:
        A null `ReceptionAlgorithm` that outputs an empty `Reception` dictionary.

    Example:
        ```
        >>> rx = get_null_reception_algorithm()
        >>> signal = torch.zeros([2, 5], dtype=torch.complex64)
        >>> rx(signal)
        {}
        ```

    """

    def reception_algorithm(signal: torch.Tensor) -> Reception:  # noqa: ARG001
        return {}

    return reception_algorithm



# ruff: noqa: D103

import time

import pytest
import torch

from torchradio import (
    Receiver,
    Reception,
    ReceptionAlgorithm,
    Transmission,
    TransmissionAlgorithm,
    Transmitter,
)
from torchradio.env import NullEnvironment
from torchradio.position import get_null_distribution


@pytest.fixture
def n_timesteps() -> int:
    return 100


@pytest.fixture
def transmission_algorithms() -> tuple[TransmissionAlgorithm, TransmissionAlgorithm]:
    def transmit1(n_timesteps: int, batch_size: int = 1) -> Transmission:
        return Transmission(
            signal=(1 + 1j)
            * torch.ones(batch_size, n_timesteps, dtype=torch.complex64),
            metadata={"time": torch.tensor(time.time())},
        )

    def transmit2(n_timesteps: int, batch_size: int = 1) -> Transmission:
        vector = (-1 - 1j) * torch.ones(batch_size, n_timesteps, dtype=torch.complex64)
        vector[
            :,
            torch.arange(n_timesteps) % 2 == 0,
        ] = 0  # map every second element to zero
        return Transmission(signal=vector, metadata={"time": torch.tensor(time.time())})

    return transmit1, transmit2


@pytest.fixture
def reception_algorithms() -> tuple[ReceptionAlgorithm, ReceptionAlgorithm]:
    def receive1(signal: torch.Tensor) -> Reception:
        # returns the sum of the inputs
        return {"sum": torch.sum(signal)}

    def receive2(signal: torch.Tensor) -> Reception:
        # returns the max and the argmax of the inputs
        mag = torch.abs(signal)
        return {"max": torch.max(mag), "argmax": torch.argmax(mag)}

    return receive1, receive2


def test_simple(
    n_timesteps: int,
    transmission_algorithms: tuple[TransmissionAlgorithm, TransmissionAlgorithm],
    reception_algorithms: tuple[ReceptionAlgorithm, ReceptionAlgorithm],
) -> None:
    transmit1, transmit2 = transmission_algorithms
    receive1, receive2 = reception_algorithms

    env = NullEnvironment()

    transmitters = {
        "tx1": Transmitter(transmit1, get_null_distribution()),
        "tx2": Transmitter(transmit2, get_null_distribution(x=10, y=10, z=10), 0.5),
    }

    receivers = {
        "rx1": Receiver(receive1, get_null_distribution(x=100)),
        "rx2": Receiver(receive2, get_null_distribution(x=50, y=80, z=10)),
    }

    env.reset()
    env.place(transmitters, receivers)
    env.simulate(n_timesteps)



"""Common helper functions for working with algorithms."""

from collections.abc import Generator, Iterable

import torch

from torchradio import Receiver, Transmitter


def get_random_bits(n_bits: int, batch_size: int = 1) -> torch.Tensor:
    """Output a random packet of bits.

    Args:
        n_bits: How many bits to output per batch.
        batch_size: How many batches to create.

    Returns:
        An 0-1 integer valued tensor with shape `[batch_size, n_bits]`.

    """
    return torch.randint(0, 2, size=[batch_size, n_bits])


def get_all_parameters(
    transmitters: Iterable[Transmitter],
    receivers: Iterable[Receiver],
) -> Generator[torch.nn.Parameter, None, None]:
    """Get the parameters for multiple devices to enable joint optimization.

    The parameters must be defined as a single iterator to be used with
    `torch.optim`.


    Args:
        transmitters: Transmitters to get parameters from.
        receivers: Receivers to get parameters from.

    Returns:
        Parameters for all devices as a single iterator.

    """
    for transmitter in transmitters:
        for p in transmitter.parameters():
            yield p

    for receiver in receivers:
        for p in receiver.parameters():
            yield p



"""Common digital modulation algorithms."""

import numpy as np
import torch
from commpy import PSKModem, QAMModem  # type: ignore

from torchradio import Reception, Transmission
from torchradio.algorithm.utils import get_random_bits
from torchradio.types import COMPLEX_DTYPE


class Modem:
    """Torchradio compatible wrapper for commpy.PSKModem and commpy.QAMModem.

    Currently implemented using the third-party `commpy` library. Note that `Modem` does not contain
    any trainable parameters. A `Modem` provides compatible `tx` and `rx` methods. When defining
    transmitters and receivers, ensure the corresponding method is passed.

    Example:
        ```
        >>> modem = Modem("psk", 8)
        >>> env = NullEnvironment()
        >>> env.place({"tx": modem.tx}, {"rx": modem.rx})
        ```

    """

    def __init__(
        self,
        mode: str,
        n_symbols: int,
        demod_type: str = "hard",
    ) -> None:
        """Create a new `Modem` using either phase-shift keying (PSK) or quadrature amplitude modulation (QAM).

        Args:
            mode: Modulation mode. Must be either "psk" or "qam".
            n_symbols: Number of symbols to use for modulation. If `mode == "psk", `n_symbols`
                must be a power of 2. If `mode == "qam"`, `n_symbols` must be a power of 2 and
                square.
            demod_type: "hard" or "soft" decision boundary for bit demodulation.


        Raises:
            ValueError: If `mode not in ["psk", "qam"]`.
            ValueError: If `n_symbols` is not a power of 2 or square (square only required for `mode == "qam"`).
            ValueError: If `demod_type not in ["hard", "soft"]`.
            NotImplementedError: If `demod_type == "soft"`.

        """
        if mode not in ["psk", "qam"]:
            err = f'{mode=} not in ["psk", "qam"]'
            raise ValueError(err)

        if demod_type not in ["hard", "soft"]:
            err = f'{demod_type=} not in ["hard", "soft"]'
            raise ValueError(err)

        if demod_type == "soft":
            raise NotImplementedError

        self._modem = PSKModem(m=n_symbols) if mode == "psk" else QAMModem(m=n_symbols)
        self._demod_type = demod_type

    @property
    def n_input_bits(self) -> int:
        """Get number of bits per symbol.

        Returns
            Number of bits per symbol.

        """
        return int(np.log2(len(self._modem.constellation)))

    def tx(self, n_timesteps: int, batch_size: int = 1) -> Transmission:
        """Apply digital modulation to randomly generated bits.

        Args:
            n_timesteps: How many timesteps to transmit for.
            batch_size: How many transmissions to create in parallel.

        Returns:
            A `Transmission` with a complex-valued signal tensor with shape `[n_timesteps, batch_size]`
                and a `metadata` dictionary that records the transmitted bits using the key "bits".

        """
        bits = get_random_bits(n_timesteps * self.n_input_bits, batch_size)
        signal = self._modulate(bits)

        return Transmission(
            signal=signal,
            metadata={"bits": bits},
        )

    def _modulate(self, bits: torch.Tensor) -> torch.Tensor:
        signal = [self._modem.modulate(b) for b in bits]
        return torch.tensor(np.array(signal), dtype=COMPLEX_DTYPE)

    def rx(self, signal: torch.Tensor) -> Reception:
        """Demodulates an input signal using a prespecified digital demodulation technique.

        Args:
            signal: A 2D complex valued input tensor.

        Returns:
            A `Reception` dictionary that records the decoded bits using the key "bits".

        """
        return {"bits": self._demodulate(signal)}

    def _demodulate(self, signal: torch.Tensor) -> torch.Tensor:
        return torch.tensor(
            np.array(
                [
                    self._modem.demodulate(s, demod_type=self._demod_type)
                    for s in signal
                ],
            ),
        )


class DSSS(Modem):
    """Direct-sequence spread spectrum modulation and demodulation.

    Currently implemented using the third-party `commpy` library. Note that `DSSS` does not contain
    any trainable parameters. `DSSS` provides compatible `tx` and `rx` methods. When defining
    transmitters and receivers, ensure the corresponding method is passed.

    Example:
        ```
        >>> chip_sequence = torch.randint(0, 2, (4,))  # binary sequence of length 4
        >>> modem = DSSS(chip_sequence)
        >>> env = NullEnvironment()
        >>> env.place({"tx": modem.tx}, {"rx": modem.rx})
        ```

    """

    _threshold = 0.5

    def __init__(
        self,
        chip_sequence: torch.Tensor,
        mode: str = "psk",
        n_symbols: int = 2,
        demod_type: str = "hard",
    ) -> None:
        """Create a new `DSSS` radio using either phase-shift keying (PSK) or quadrature amplitude modulation (QAM).

        Args:
            chip_sequence: A binary sequence to spread the input bits with.
            mode: Modulation mode. Must be either "psk" or "qam".
            n_symbols: Number of symbols to use for modulation. If `mode == "psk", `n_symbols`
                must be a power of 2. If `mode == "qam"`, `n_symbols` must be a power of 2 and
                square.
            demod_type: "hard" or "soft" decision boundary for bit demodulation.


        Raises:
            ValueError: If `mode not in ["psk", "qam"]`.
            ValueError: If `n_symbols` is not a power of 2 or square (square only required for `mode == "qam"`).
            ValueError: If `demod_type not in ["hard", "soft"]`.
            NotImplementedError: If `demod_type == "soft"`.

        """
        super().__init__(mode, n_symbols, demod_type)
        self._chip_sequence = chip_sequence

    def tx(self, n_timesteps: int, batch_size: int = 1) -> Transmission:
        """Encode randomly generated bits with a chip sequence before modulating using commpy.Modem.

        Args:
            n_timesteps: How many timesteps to transmit for.
            batch_size: How many transmissions to create in parallel.

        Returns:
            A `Transmission` with a complex-valued signal tensor with shape `[n_timesteps, batch_size]`
                and a `metadata` dictionary that records the transmitted bits using the key "bits".

        """
        n_bits = n_timesteps * self.n_input_bits / len(self._chip_sequence)

        if not n_bits.is_integer():
            err = f"A chip sequence of length {len(self._chip_sequence)} for {n_timesteps} timesteps requires {n_bits} bits. Please reselect n_timesteps such that n_timesteps * self.n_input_bits / len(self._chip_sequence) is an integer."
            raise ValueError(err)

        bits = get_random_bits(int(n_bits), batch_size)

        repeated_bits = torch.repeat_interleave(bits, len(self._chip_sequence), dim=-1)
        tiled_chip_sequence = torch.tile(self._chip_sequence, [batch_size, int(n_bits)])
        dsss = torch.logical_xor(repeated_bits, tiled_chip_sequence)

        return Transmission(
            signal=self._modulate(dsss),
            metadata={"bits": bits},
        )

    def rx(self, signal: torch.Tensor) -> Reception:
        """Demodulates and despreads an input signal.

        Args:
            signal: A 2D complex valued input tensor.

        Returns:
            A `Reception` dictionary that records the decoded bits using the key "bits".

        """
        batch_size = signal.shape[0]
        dsss = self._demodulate(signal)

        return {
            "bits": (
                torch.logical_xor(
                    dsss.reshape(batch_size, -1, len(self._chip_sequence)),
                    self._chip_sequence,
                )
                .type(torch.float)
                .mean(axis=-1)
                > self._threshold
            )
            .type(torch.int)
            .reshape(batch_size, -1),
        }



# ruff: noqa: D103

from torchradio.algorithm.example import DenseRadio


def test_example() -> None:
    dense_radio = DenseRadio(8, 2)
    tx = dense_radio.tx(64, 2)
    rx = dense_radio.rx(tx.signal)

    assert tx.metadata["bits"].shape == rx["bits"].shape



from torchradio.device import Position, Receiver, SpatialDistribution, Transmitter
from torchradio.types import (
    DeviceLogs,
    Reception,
    ReceptionAlgorithm,
    Transmission,
    TransmissionAlgorithm,
)

__all__ = [
    "Transmission",
    "Reception",
    "TransmissionAlgorithm",
    "ReceptionAlgorithm",
    "Transmitter",
    "Receiver",
    "SpatialDistribution",
    "Position",
    "DeviceLogs",
]



"""Abstract environments that do not require physical device placements."""

import torch

from torchradio.device import Position
from torchradio.env.base import BaseEnvironment


class NullEnvironment(BaseEnvironment):
    """The simplest possible environment with ideal conditions."""

    def _compute_propagation_parameters(self) -> None:
        """No propagation parameters to compute."""

    def _in_bounds(self, position: Position) -> bool:  # noqa: ARG002
        """Everything is in bounds."""
        return True

    def _propagate(self, tx: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        """No attenuation or reflection. Simply sum all transmissions together."""
        summed = torch.sum(torch.stack(list(tx.values()), axis=1), axis=1)  # type: ignore
        return {receiver_name: summed for receiver_name in self._receivers}

    def _get_background_noise(
        self,
        n_timesteps: int,
        batch_size: int = 1,
    ) -> dict[str, torch.Tensor]:
        """No background noise."""
        return {
            receiver_name: torch.zeros([batch_size, n_timesteps], dtype=torch.complex64)
            for receiver_name in self._receivers
        }


class ControlledSNREnvironment(NullEnvironment):
    """A single-channel environment that applies additive white Gaussian noise (AWGN) to a specific signal-to-noise ratio (SNR).

    A null environment with controllable SNR that supports a single transmitter and receiver pair.
    During propagation, the transmitted signal is reduced to unit power. The correct amount of AWGN is
    then generated to achieve the specified SNR.

    Raises
        ValueError: If more than one transmitter or more than one receiver.

    """

    def __init__(self, snr: float) -> None:
        """Create a `ControlledSNREnvironment` for single channel simulations with Gaussian noise.

        Args:
            snr: Desired signal-to-noise ratio in dB.

        """
        self._snr = torch.tensor([snr])

    @property
    def snr(self) -> float:
        """Get the currently set signal-to-noise ratio in dB.

        Returns
            Current signal-to-noise ratio in dB.

        """
        return self._snr.numpy()[0]

    def set_snr(self, snr: float) -> None:
        """Set signal-to-noise ratio (SNR) for the receiver in dB.

        Args:
            snr: Desired signal-to-noise ratio in dB.

        """
        self._snr = torch.tensor([snr])

    def _compute_propagation_parameters(self) -> None:
        """No propagation parameters to compute - check that only one transmitter and one receiver have been placed."""
        if len(self.transmitters) != 1:
            err = f"{self.__class__.__name__} only supports the placement of a single transmitter. {len(self.transmitters)=} != 1"
            raise ValueError(err)

        if len(self.receivers) != 1:
            err = f"{self.__class__.__name__} only supports the placement of a single receiver. {len(self.receivers)=} != 1"
            raise ValueError(err)

    def _propagate(self, tx: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        """Attenuate signal to low power such that, when it is summed with unit power background noise, the desired SNR is achieved."""
        summed = torch.sum(torch.stack(list(tx.values()), axis=1), axis=1)  # type: ignore
        norm_tx = summed / torch.sqrt(torch.var(summed))
        tx_ = torch.sqrt(10 ** (self._snr / 10)) * norm_tx
        return {receiver_name: tx_ for receiver_name in self._receivers}

    def _get_background_noise(
        self,
        n_timesteps: int,
        batch_size: int = 1,
    ) -> dict[str, torch.Tensor]:
        """Get background noise with unit power."""
        noise = torch.randn(batch_size, n_timesteps) + 1j * torch.randn(
            batch_size,
            n_timesteps,
        )
        norm_noise = noise / torch.sqrt(torch.var(noise))
        return {receiver_name: norm_noise for receiver_name in self._receivers}


class RandomBoundedSNREnvironment(ControlledSNREnvironment):
    """A single-channel environment that applies a random amount of additive white Gaussian noise (AWGN) according to preset SNR bounds.

    A null environment with bounded SNR that supports a single transmitter and receiver pair.
    During propagation, the transmitted signal is reduced to unit power. The correct amount of AWGN is
    then generated to achieve an SNR within the provided bounds.

    Raises
        ValueError: If more than one transmitter or more than one receiver.

    """

    def __init__(self, snr_min: float, snr_max: float) -> None:
        """Create a `RandomBoundedSNREnvironment` for single channel simulations with Gaussian noise.

        Args:
            snr_min: Minimum desired signal-to-noise ratio in dB.
            snr_max: Maximum desired signal-to-noise ratio in dB.

        Raise:
            ValueError: if `snr_max < snr_min`.

        """
        if snr_max < snr_min:
            err = f"{snr_max=}<{snr_min=}."
            raise ValueError(err)

        self._snr_min = snr_min
        self._snr_diff = snr_max - snr_min

    def _propagate(self, tx: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        """Attenuate signal to low power such that, when it is summed with unit power background noise, the desired SNR is achieved."""
        snr = self._snr_min + torch.rand(1) * self._snr_diff
        summed = torch.sum(torch.stack(list(tx.values()), axis=1), axis=1)  # type: ignore
        norm_tx = summed / torch.sqrt(torch.var(summed))
        tx_ = torch.sqrt(10 ** (snr / 10)) * norm_tx
        return {receiver_name: tx_ for receiver_name in self._receivers}


class RandomAWGNEnvironment(NullEnvironment):
    """A null environment that produces a random level of background noise for each receiver.

    This random level is sampled per receiver according to the power bounds supplied at instantiation.
    """

    def __init__(
        self,
        p_min: float,
        p_max: float,
        *,
        normalize_before_noise: bool = False,
    ) -> None:
        """Initialize a new `RandomAWGNEnviroment` with bounded background noise levels.

        Args:
            p_min: Minimum background noise power.
            p_max: Maximum background noise power.
            normalize_before_noise: Set to True if the aggregated signals should be normalized
                to unit power before adding background noise. Defaults to False.

        """
        self.set_bounds(p_min, p_max)
        self._normalize = normalize_before_noise

    @property
    def bounds(self) -> tuple[float, float]:
        """Get current noise bounds.

        Returns
            Background noise bounds as a tuple.

        """
        return self._p_min, self._p_max

    def set_bounds(self, p_min: float, p_max: float) -> None:
        """Set noise bounds for all receivers.

        Args:
            p_min: Minimum background noise power.
            p_max: Maximum background noise power.

        """
        if p_max < p_min:
            err = f"{p_max=} < {p_min=}"
            raise ValueError(err)

        if p_min < 0:
            err = f"{p_min=} < 0"

        self._p_min = p_min
        self._p_max = p_max

    def _propagate(self, tx: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        """Randomly attenuate each transmission and sum.

        The amount of random attenutation is different for each transmitter/receiver pair.
        If self.normalize == True, the received signals are normalized to unit power.
        """
        # copy each tx for each rx
        concatenated = torch.concat([tx_[None] for tx_ in tx.values()])
        duplicated = torch.tile(
            concatenated,
            [self.n_receivers, 1, 1, 1],
        )  # [rx, tx, batch, t]

        # take weighted sum
        weights = torch.rand(self.n_receivers, len(tx), duplicated.shape[2])
        weights /= weights.sum(axis=1, keepdim=True)  # average over transmitters
        weighted = weights[..., None] * duplicated
        result = weighted.sum(axis=1)  # [rx, batch, t]

        if self._normalize:
            powers = torch.var(result, axis=-1)
            result = result / torch.sqrt(powers)

        return dict(zip(self._receivers, result, strict=True))

    def _get_background_noise(
        self,
        n_timesteps: int,
        batch_size: int = 1,
    ) -> dict[str, torch.Tensor]:
        """Get background noise with random power level for each batch/receiver sampled in self.bounds."""
        # Get background noise with unit power
        noise = torch.randn(
            self.n_receivers,
            batch_size,
            n_timesteps,
        ) + 1j * torch.randn(
            self.n_receivers,
            batch_size,
            n_timesteps,
        )
        norm_noise = noise / torch.sqrt(torch.var(noise, axis=-1))[..., None]

        # Apply random power level for each batch/receiver
        noise_levels = self._p_min + (self._p_max - self._p_min) * torch.rand(
            self.n_receivers,
            batch_size,
        )
        adjusted_noise = torch.sqrt(noise_levels[..., None]) * norm_noise

        return dict(zip(self._receivers, adjusted_noise, strict=True))



# ruff: noqa: D103

import numpy as np
import pytest
import torch

from torchradio import Receiver, Transmitter
from torchradio.algorithm import Modem
from torchradio.env.null import ControlledSNREnvironment
from torchradio.position import get_null_distribution


@pytest.fixture
def n_timesteps() -> int:
    return 100


@pytest.fixture
def modem() -> Modem:
    return Modem("psk", 2)


@pytest.fixture
def transmitter(modem: Modem) -> Transmitter:
    return Transmitter(modem.tx, get_null_distribution())


@pytest.fixture
def receiver(modem: Modem) -> Receiver:
    return Receiver(modem.rx, get_null_distribution())


@pytest.fixture
def env() -> ControlledSNREnvironment:
    return ControlledSNREnvironment(0)


@pytest.mark.parametrize("snr", [-10.0, -5.0, 0.0, 5.0, 10.0])
def test_snr(
    snr: float,
    env: ControlledSNREnvironment,
    n_timesteps: int,
    transmitter: Transmitter,
    receiver: Receiver,
) -> None:
    """Check that ControlledSNREnvironment produces signals at the receiver with the correct SNR."""
    transmitter_name = "test_tx"
    receiver_name = "test_rx"
    transmitters = {transmitter_name: transmitter}
    receivers = {receiver_name: receiver}

    env.reset()
    env.set_snr(snr)
    env.place(transmitters, receivers)
    device_logs = env.simulate(n_timesteps)

    # check that transmission is a raw multiple of (received signal minus background noise)
    transmission = device_logs.tx[transmitter_name].signal
    background_noise = device_logs.rx[receiver_name]["noise"]
    propagated = device_logs.rx[receiver_name]["raw"] - background_noise
    ratios = torch.abs(transmission) / torch.abs(propagated)
    assert torch.allclose(ratios[0], ratios)

    # check SNR is correct
    signal_power = torch.var(propagated).numpy()
    noise_power = torch.var(background_noise).numpy()
    computed_snr = 10 * np.log10(signal_power / noise_power)
    assert np.isclose(snr, computed_snr, atol=1e-3)


def test_single_pair(
    env: ControlledSNREnvironment,
    transmitter: Transmitter,
    receiver: Receiver,
) -> None:
    """Check that ControlledSNREnvironment raises a ValueError if more than one transmitter or receiver is place."""
    transmitters = {"a": transmitter, "b": transmitter}
    receivers = {"a": receiver, "b": receiver}

    env.reset()
    with pytest.raises(
        ValueError,
        match="only supports the placement of a single transmitter",
    ):
        env.place(transmitters, receivers)



"""Defines the core types used throughout Torchradio."""

from dataclasses import dataclass
from typing import Protocol

import torch

COMPLEX_DTYPE = torch.complex64
""" The common datatype for all complex tensors in Torchradio."""


@dataclass
class Transmission:
    """The core output type for any transmitter. A `Transmission` consists of two parts: a signal and a metadata dictionary.

    Args:
        signal: A complex-valued tensor that is propagated into the environment
        metadata: Used to record arbitrary information that can be used by downstream loss functions.
            The most commonly used metadata key is 'bits' to capture the bits before the `TransmissionAlgortihm`
            produces an output signal.

    """

    signal: torch.Tensor
    metadata: dict[str, torch.Tensor]


Reception = dict[str, torch.Tensor]
""" The `ReceptionAlgorithm` author is free to capture whatever fields are required for downstream loss functions. A `Reception` is
the receiver analogue of `Transmission.metadata`."""


@dataclass
class DeviceLogs:
    """Captures the `Transmission` and `Reception` logs from a simulation run.

    Each `Transmission` and `Reception` is associated with a device name.

    Args:
        tx: Maps device names to `Transmission`s.
        rx: Maps device names to `Reception`s.

    """

    tx: dict[str, Transmission]
    rx: dict[str, Reception]


class TransmissionAlgorithm(Protocol):
    """All transmitters in Torchradio are required to implement a `TransmissionAlgorithm`.

    A `TransmissionAlgorithm` is any `Callable` that produces a `Transmission` with a complex-valued
    `signal` with shape `[batch_size, n_timesteps]`.
    """

    def __call__(self, n_timesteps: int, batch_size: int) -> Transmission:
        """Return a `Transmission` given `n_timesteps` and `batch_size`.

        Args:
            n_timesteps: Number of timesteps to simulate.
            batch_size: How many batches to yield.

        Returns:
            A `Transmission` with a complex-valued signal with shape `[batch_size, n_timesteps]`.

        """
        ...


class ReceptionAlgorithm(Protocol):
    """All receivers in Torchradio are required to implement a `ReceptionAlgorithm`.

    A `ReceptionAlgorithm` is any `Callable` that produces a `Reception` from an input `signal`.

    """

    def __call__(self, signal: torch.Tensor) -> Reception:
        """Return a `Reception` from a complex-valued input `signal`.

        Args:
            signal: A complex-valued input tensor from the environment.

        Returns:
            A `Reception` with that maps each key to a `torch.Tensor` for downstream loss functions.

        """
        ...



"""Groups together common classes and functions for randomly sampling physical device positions in environments."""

from dataclasses import dataclass
from typing import Protocol

import torch


@dataclass
class Position:
    """A concrete instantiation of a device's position in the environment.

    Args:
        x: x coordinate
        y: y coordinate
        z: z coordinate

    """

    x: float
    y: float
    z: float


class SpatialDistribution(Protocol):
    """All devices are required to include a `SpatialDistribution`.

    A `SpatialDistribution` is any `Callable` returns a randomly sampled `Position` each time it is invoked.

    """

    def __call__(self) -> Position:
        """Randomly sample a `Position` from an underlying distribution.

        Returns
            A randomly sampled `Position`.

        """
        ...


def get_uniform_distribution(
    x_bounds: tuple[float, float],
    y_bounds: tuple[float, float],
    z_bounds: tuple[float, float],
) -> SpatialDistribution:
    """Get a `SpatialDistribution` that samples uniformly from some 3D bounds.

    Args:
        x_bounds: Desired interval for the x coordinate
        y_bounds: Desired interval for the y coordinate
        z_bounds: Desired interval for the z coordinate

    Returns:
        A `SpatialDistribution` that returns a uniformly-sampled `Position` according to the provided bounds.

    Example:
        ```
        >>> distribution = get_uniform_distribution((0, 10), (0, 10), (0, 5))
        >>> distribution()
        Position(x=9.82526, y=1.6853619, z=1.1326883)
        ```

    Raises:
        ValueError: If `bounds[1] < bounds[0]`.

    """
    x_min, x_max = x_bounds
    y_min, y_max = y_bounds
    z_min, z_max = z_bounds

    if x_max < x_min:
        err = f"{x_max=}<{x_min=}"
        raise ValueError(err)

    if y_max < y_min:
        err = f"{y_max=}<{y_min=}"
        raise ValueError(err)

    if z_max < z_min:
        err = f"{z_max=}<{z_min=}"
        raise ValueError(err)

    min_tensor = torch.tensor([x_min, y_min, z_min])
    max_tensor = torch.tensor([x_max, y_max, z_max])
    diff_tensor = max_tensor - min_tensor

    def distribution() -> Position:
        as_tensor = min_tensor + (torch.rand(3) * diff_tensor)
        return Position(*as_tensor.numpy())

    return distribution


def get_null_distribution(
    x: float = 0,
    y: float = 0,
    z: float = 0,
) -> SpatialDistribution:
    """Get a trivial or (null) `SpatialDistribution` that always samples the same `Position`.

    Args:
        x: x coordinate
        y: y coordinate
        z: z coordinate

    Returns:
        A `SpatialDistribution` that always returns the same `Position` as described by `x`, `y` and `z`.

    Example:
        ```
        >>> distribution = get_null_distribution(1, 2)
        >>> distribution()
        Position(x=1, y=2, z=0)
        ```

    """
    return lambda: Position(x, y, z)



"""Defines the `BaseEnvironment` that all `Torchradio` environments are expected to inherit from."""

import warnings
from abc import ABC, abstractmethod
from os import PathLike

import torch

from torchradio.algorithm.null import (
    get_constant_transmission_algorithm,
    get_null_reception_algorithm,
)
from torchradio.device import Position, Receiver, Transmitter
from torchradio.types import DeviceLogs


class BaseEnvironment(ABC):
    """Abstract base class for `Torchradio` environments.

    The core concept behind `Torchradio` is the use of a common simulation environment
    that enables backpropagation from the loss function back to the transmitters. It is
    assumed that all environments used with `Torchradio` are child classes of
    `BaseEnvironment`. The `BaseEnvironment` automates many common operations, such as
    device placement, noise simulation and signal aggregation.

    All child classes of `BaseEnvironment` are expected to implement the abstract methods:
    `_compute_propagation_parameters`, `_in_bounds`, `_propagate` and `_get_background_noise`.
    """

    def __init__(self, *, disable_differentiability_check: bool = False) -> None:
        """Create a clean environment that can be populated with devices."""
        if not disable_differentiability_check and not self.is_differentiable():
            msg = f"{self} does not appear to be differentiable. Check your environment logic. Raise an issue at the GitHub repository if you have received this warning in error."
            warnings.warn(msg, stacklevel=2)
        self.reset()

    def reset(self) -> None:
        """Remove all devices from the environment.

        Example:
            ```
            >>> env.place(...)
            >>> env.n_devices
            8
            >>> env.reset()
            >>> env.n_devices
            0
            ```

        """
        self._transmitters: dict[str, Transmitter] = {}
        self._receivers: dict[str, Receiver] = {}

        self._transmitter_positions: dict[str, Position] = {}
        self._receiver_positions: dict[str, Position] = {}

    @property
    def n_transmitters(self) -> int:
        """Get current number of transmitters.

        Returns
            The number of transmitters placed in the environment.

        """
        return len(self._transmitters)

    @property
    def n_receivers(self) -> int:
        """Get current number of receivers.

        Returns
            The number of receivers placed in the environment.

        """
        return len(self._receivers)

    @property
    def n_devices(self) -> int:
        """Get current number of devices.

        Returns
            The number of devices placed in the environment.

        """
        return self.n_transmitters + self.n_receivers

    @property
    def transmitters(self) -> dict[str, Position]:
        """Get current transmitters and their positions.

        Returns
            A map from transmitter names to positions

        """
        return self._transmitter_positions

    @property
    def receivers(self) -> dict[str, Position]:
        """Get current receivers and their positions.

        Returns
            A map from receiver names to positions

        """
        return self._receiver_positions

    @property
    def devices(self) -> dict[str, dict[str, Position]]:
        """Get a dictionary that summarizes the environment's devices with their current positions.

        Returns
            Two maps. One that maps transmitter names to positions and one that maps receiver names to positions.

        """
        return {"transmitters": self.transmitters, "receivers": self.receivers}

    def place(
        self,
        transmitters: dict[str, Transmitter],
        receivers: dict[str, Receiver],
    ) -> None:
        """Place devices in the environment.

        Child classes must specify whether a device is in or out-of-bounds via the `_in_bounds` method.
        After the devices have been placed, `self._compute_propagation_parameters()` is called to determine
        simulation parameters. These simulation parameters do not need to be recomputed until the devices have
        been re-placed.

        Args:
            transmitters: Maps device names to `Transmitter`s.
            receivers: Maps device names to `Receivers`s.

        Example:
            ```
            >>> transmitter_1 = torchradio.algorithm.null.get_constant_transmitter(1 + 1j)
            >>> transmitter_2 = torchradio.algorithm.null.get_constant_transmitter(1 + 0j)
            >>> transmitter_3 = torchradio.algorithm.null.get_null_transmitter()
            >>> receiver = torchradio.algorithm.null.get_null_receiver()
            >>> some_transmitters = {"tx1": transmitter_1, "tx2": transmitter_2}
            >>> other_transmitters = {"tx3": transmitter_3}
            >>> all_transmitters = {**some_transmitters, **other_transmitters}
            >>> receivers = {"rx": receiver}
            >>> env = torchradio.env.null.NullEnvironment()
            >>> env.place(some_transmitters, receivers)
            >>> env.n_devices
            3
            >>> env.place(all_transmitters, receivers)
            >>> env.n_devices
            4
            ```

        Raises:
            ValueError: No receivers provided.
            TypeError: A non-`Transmitter` was found in `transmitters` or a non-`Receiver` was found in `receivers`.

        """
        if len(receivers) == 0:
            err = f"{self.__class__.__name__} requires at least one receiver: {len(receivers)=}"
            raise ValueError(err)

        for transmitter_name, transmitter in transmitters.items():
            if not isinstance(transmitter, Transmitter):
                err = f"transmitters contains a non-transmitter: {transmitter_name}: {transmitter.__class__.__name__}"
                raise TypeError(err)

        for receiver_name, receiver in receivers.items():
            if not isinstance(receiver, Receiver):
                err = f"receivers contains a non-receiver: {receiver_name}: {receiver.__class__.__name__}"
                raise TypeError(err)

        self.reset()

        self._transmitters = transmitters
        self._receivers = receivers

        self._transmitter_positions = self._place_devices(self._transmitters)
        self._receiver_positions = self._place_devices(self._receivers)

        self._compute_propagation_parameters()

    def simulate(self, n_timesteps: int, batch_size: int = 1) -> DeviceLogs:
        """Run the simulation for `n_timesteps` with `batch_size`.

        Args:
            n_timesteps: How many timesteps to simulate. Must be positive.
            batch_size: How many batches to simulate. Must be positive.

        Returns:
            Device logs for benchmarking performance and computing gradients.

        Example:
            ```
            >>> transmitter = torchradio.algorithm.null.get_null_transmitter()
            >>> receiver = torchradio.algorithm.null.get_null_receiver()
            >>> env = torchradio.env.null.NullEnvironment()
            >>> env.place({"tx": transmitter}, {"rx": receiver})
            >>> env.simulate(20, 3)
            ```

        Raises:
            ValueError: If a provided `SpatialDistribution` is incompatible with the environment
                according to the child class `_in_bounds` methods.
            RuntimeError: If child classes don't correctly implement `_propagate` or
                `_get_background_noise` to account for all devices.

        """
        # get transmissions from every transmitter in the environment for n_timesteps
        tx_logs = {}
        for transmitter_name, transmitter in self._transmitters.items():
            transmission = transmitter(n_timesteps, batch_size)

            if tuple(transmission.signal.shape) != (batch_size, n_timesteps):
                err = f"Received signal with shape {transmission.signal.shape} from transmitter: {transmitter_name}. Expected {(batch_size, n_timesteps)}."
                raise RuntimeError(err)

            tx_logs[transmitter_name] = transmission

        # propagate transmitted signals to every receiver in the environment
        if self.n_transmitters > 0:
            propagations = self._propagate(
                {
                    transmitter_name: tx.signal
                    for transmitter_name, tx in tx_logs.items()
                },
            )
            if propagations.keys() != self._receivers.keys():
                err = f"self._propagate did not account for all receivers: {propagations.keys()=} vs {self._receivers.keys()=}."
                raise RuntimeError(err)
        else:
            propagations = {
                receiver_name: torch.zeros(
                    [batch_size, n_timesteps],
                    dtype=torch.complex64,
                )
                for receiver_name in self.receivers
            }

        # get random background noise for every receiver in the environment
        background_noise = self._get_background_noise(n_timesteps, batch_size)
        if background_noise.keys() != self._receivers.keys():
            err = f"self._get_background_noise did not account for all receivers: {background_noise.keys()=} vs {self._receivers.keys()=}."
            raise RuntimeError(err)

        # sum the aggregated transmissions with the background noise for every receiver in the environment
        rx = {
            receiver_name: background_noise[receiver_name] + propagations[receiver_name]
            for receiver_name in self._receivers
        }

        # process the received signals
        rx_logs = {}
        for receiver_name, receiver in self._receivers.items():
            raw = rx[receiver_name]
            rx_logs[receiver_name] = receiver(raw)

            # track received signal before receiver processing
            if "raw" in rx_logs[receiver_name]:
                err = f'{receiver_name} uses the reserved keyword "raw" in the Reception dictionary.'
                raise RuntimeError(err)
            rx_logs[receiver_name]["raw"] = raw

            # record isolated background noise for training and analysis
            if "noise" in rx_logs[receiver_name]:
                err = f'{receiver_name} uses the reserved keyword "noise" in the Reception dictionary.'
                raise RuntimeError(err)
            rx_logs[receiver_name]["noise"] = background_noise[receiver_name]

        return DeviceLogs(tx=tx_logs, rx=rx_logs)

    def _place_devices(
        self,
        devices: dict[str, Transmitter] | dict[str, Receiver],
        max_n_attempts: int = 100,
    ) -> dict[str, Position]:
        """Place an iterable of devices in the environment according to their spatial distributions.

        Args:
            devices: Maps devices names to `Device`s.
            max_n_attempts: How many attempts to try placing each device before raising a `ValueError`.

        Returns:
            A mapping from device names to `Position`s.

        Raises:
            ValueError: If a device cannot be suitably placed within `max_n_attempts`. This indicates
                that the provided `SpatialDistribution` is unsuitable for the current environment.

        """
        positions = {}
        for device_name, device in devices.items():
            n_attempts = 0
            while True:
                n_attempts += 1
                if n_attempts >= max_n_attempts:
                    err = f"Device {device_name} has a spatial distribution that is incompatible with the current environment."
                    raise ValueError(err)

                position = device.place()

                if self._in_bounds(position):
                    break

            positions[device_name] = position

        return positions

    def visualize(
        self,
        *,
        show: bool = True,
        save_path: PathLike | None = None,
    ) -> None:
        """Visualize the environment with the currently placed devices.

        It is not mandatory for child classes to override this method. However, it may make it
        easier for users to interact with the environment if they are provided with a convenient
        visualization method to see where devices are currently placed.

        Args:
            show: Show an interactive plot within the Python process.
            save_path: Path to save the visualization to.

        """
        raise NotImplementedError

    @abstractmethod
    def _compute_propagation_parameters(self) -> None:
        """Compute and store any necessary propagation parameters as object attributes before the simulation begins."""
        ...

    @abstractmethod
    def _in_bounds(self, position: Position) -> bool:
        """Determine whether the device is in bounds according to the environment's specifications.

        Args:
            position: A position in the environment

        Returns:
            True if the position is valid according to the environment's specifications.

        """
        ...

    @abstractmethod
    def _propagate(self, tx: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        """Propagate transmitted signals to every receiver in the environment.

        Args:
            tx: A map from transmitter names to transmitted signals.

        Returns:
            A dictionary mapping each receiver to a complex-valued input signal (not including background noise).

        """
        ...

    @abstractmethod
    def _get_background_noise(
        self,
        n_timesteps: int,
        batch_size: int = 1,
    ) -> dict[str, torch.Tensor]:
        """Compute random background noise for every receiver in the environment.

        Args:
            n_timesteps: Number of timesteps of background noise.
            batch_size: How many batches of background noise.

        Returns:
            A dictionary mapping each receiver to a complex-valued background noise
                tensor with shape `[n_timesteps, batch_size]`.

        """
        ...

    def is_differentiable(self) -> bool:
        """Check the environment is differentiable.

        Gotcha: The differentiability check assumes that devices can be placed at
        Position(x=0, y=0, z=0). If this is impossible, you must either write your own
        differentiability check, or alter your environment to allow placements at
        the origin.

        Gotcha: As part of the differentiability check, the environment will be reset.
        If you have devices placed in the environment before calling this function, they
        will need to be re-placed afterwards.

        Returns:
            True if the environment passes a basic differentiability test, where the output
                magnitude of a simple transmitter is driven to zero. If your environment is
                incorrectly classified as non-differentiable, please raise an issue at
                the repository issue tracker.

        Example:
                ```
                >>> env = torchradio.env.null.NullEnvironment()
                >>> env.is_differentiable()
                True
                ```

        """
        self.reset()

        # get basic trainable transmitter and a null receiver
        init_val = 1 + 1j
        weight = torch.nn.Parameter(torch.tensor(init_val))
        transmitter = Transmitter(get_constant_transmission_algorithm(weight))
        receiver = Receiver(get_null_reception_algorithm())

        # place and simulate
        self.place({"tx": transmitter}, {"rx": receiver})
        device_logs = self.simulate(10, 2)

        # create an arbitrary loss function that drives the raw received signal to zero
        raw = device_logs.rx["rx"]["raw"]
        sum_squares = torch.sum(raw.real**2 + raw.imag**2)
        opt = torch.optim.SGD([weight], lr=0.1)

        try:
            torch.nn.MSELoss()(
                sum_squares,
                torch.tensor(0, dtype=torch.float32),
            ).backward()
        except RuntimeError:
            return False  # tensors does not require grad and does not have a grad_fn

        opt.step()

        self.reset()

        return (
            weight.item() != init_val
        )  # if differentiable, the weight parameter should have changed



import pytest
import torch

from torchradio.algorithm import DSSS


@pytest.mark.parametrize(
    ("chip_sequence", "mode", "n_symbols"),
    [
        (torch.tensor([1, 0]), "psk", 2),
        (torch.tensor([1, 0]), "qam", 4),
        (torch.tensor([1, 0, 1, 1]), "psk", 4),
        (torch.tensor([1, 0, 1, 1]), "qam", 16),
        (torch.tensor([1, 1, 1, 1, 0, 0, 1, 1]), "psk", 2),
        (torch.tensor([1, 0, 1, 1, 0, 1, 1, 1]), "psk", 16),
        (torch.tensor([1, 0, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1]), "qam", 16),
        (torch.tensor([1, 0, 0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1]), "psk", 2),
    ],
)
def test_dsss(chip_sequence: torch.Tensor, mode: str, n_symbols: int) -> None:
    """Check DSSS modulation and demodulation are correct."""
    dsss = DSSS(chip_sequence, mode, n_symbols)
    transmission = dsss.tx(80, 4)
    reception = dsss.rx(transmission.signal)
    assert torch.all(transmission.metadata["bits"] == reception["bits"]).item()
